\ ==============================================================
\ [TITLE]     Byte-Forth DCF77 decoder primitives. 
\ [FILE]      DCF.FRT
\ [CATEGORY]  DCF77 project file: 1177 bytes, opt: 1068 bytes
\ [VSN]       1.00 
\ [CREATED]   October 25, 1996, Willem Ouwerkerk 
\ [LASTCHNGD] November 11, 1996, Willem Ouwerkerk 
\ [COPYRIGHT] Dutch Forth Users Group (C) 1996
\ [COMPILE]   ByteForth Environments
\ [MODULE OF] ByteForth 
\ [USES]      RS232.FRT, NUMBERS.FRT
\ [AUTHOR]    Willem Ouwerkerk 
\ [PURPOSE]   Read DCF77 clock bits using bit 1 of Port P1 and
\             timer 1 to test incoming bits for true or false.
\
\             INFO byte, all bits are active high (5 bits):   
\               -0- Backup transmitter on.
\               -1- Time change summer- wintertime (one hour in front).
\               -2- Summer time active.
\               -3- Summer time timezone (always zero ?).
\               -4- Correctie tussen UTC en astronomische tijd 
\                   (one hour in front).
\
\             Weekday byte (3 bits):- 001 is monday, 111 is sunday.
\
\             BCD means that every 4 bits contains a decimal digit
\             that is a number from 0 to 9 in binary format.
\ 
\ 12 MHz code timing, besides MS that is a 6 MHz routine.
\
\ ==============================================================

89C2051  TARGET                         \ Instruct compiler
$25 $01 $20 MEMORY                      \ Set memory layout

ATOM INLINE$    ATOM UM/MOD             \ Words wanted in dictionary

\ Internal version of DCF77 registers

VARIABLE MINS:                          \ 0  Minutes !
VARIABLE HOURS:                         \ 1
VARIABLE DAYS:                          \ 2
VARIABLE MONTHS:                        \ 3
VARIABLE YEARS:                         \ 4
VARIABLE WEEKDAY:                       \ 5
VARIABLE INFO:                          \ 6

VARIABLE WORK                           \ Dcf work register
VARIABLE BITS                           \ Count DCF77 bits

\G The contents of this register is $FF when the decoder is
\G synchronised to the DCF77 signal, or $00 when not.
FLAG SYNC?                              \ High when synchronised
FLAG SEC?                               \ A second passed 
FLAG BIT                                \ Last received bit (0/1)

$B3 BIT-SFR DCF-IN                      \ Input for DCF77 signal
$92 BIT-SFR LED1                        \ Trace leds = TICK
$95 BIT-SFR LED2                        \ = STICK

\ User version of DCF77 clock registers

VARIABLE SECONDS                        \ 0 to 59 seconds
VARIABLE MINUTES                        \ 7 bits
VARIABLE HOURS                          \ 6 bits
VARIABLE DAYS                           \ 6 bits
VARIABLE MONTHS                         \ 5 bits
VARIABLE YEARS                          \ 8 bits
VARIABLE WEEKDAY                        \ 3 bits
VARIABLE INFO                           \ 5 bits

VARIABLE TICK                           \ Used for synchronisation
VARIABLE STICK                          \ Maintain Clock without DCF77 signal

$D8      =: HPERIOD                     \ Dcf Ticks interval period
$F0      =: LPERIOD                     \ About 1/100 second

    $8C =: TH0:                         \ Assembler SFR labels   
    $8A =: TL0:
    $89 =: TMOD:
    $A9 =: ET0:
    $8C =: TR0:   
    $AF =: EAL:
    $BB =: EX4:
    $99 =: SBUF:                        \ Labels for RS232
    $99 =: TI:
    $98 =: RI:
    $87 =: PCON:
    $8D =: TH1:
    $8E =: TR1:
    $98 =: SCON:
    $AC =: ES:

\ [PURPOSE]   RS232 I/O on target with use of timer 1.

CODE SETUP-RS232
    TMOD: $20 # MOV,                    \ Timer 1 mode 2 = 8 bit autoreload
    TR1: SETB,                          \ Run timer 1
    TH1: 243 # MOV,                     \ Select 2400/4800 baud
    PCON: $80 # ORL,                    \ At start at 4800 baud
    SCON: $7A # MOV,  ( $52 )           \ Serial mode 1
    ES: CLR,                            \ Disable SIO interrupt
    RET,
END-CODE

CODE RS232-EMIT     ( char -- )
    A: @R0 MOV,
    R0: INC,
    BEGIN, TI: BS UNTIL,
    SBUF: A: MOV,
    TI: CLR,
    RET,
END-CODE

CODE RS232-TYPE     ( $addr u -- )
    [R1] @R0 MOV,
    R0: INC,
    DPH: @R0 MOV,
    R0: INC,
    DPL: @R0 MOV,
    R0: INC,
    A: R1: MOV,
    NZR IF,
        DO,     A: CLR,
                A: @A+DPTR MOVC,
                BEGIN, TI: BS UNTIL,
                SBUF: A: MOV,
                TI: CLR,
                DPTR: INC,
        R1: LOOP,
    THEN,
    RET,
END-CODE

CODE RS232-RTYPE    ( addr u -- )
    [R2] @R0 MOV,
    R0: INC,
    [R1] @R0 MOV,
    R0: INC,
    A: R2: MOV,
    NZR IF,
        DO,     BEGIN, TI: BS UNTIL,
                SBUF: @R1 MOV,
                TI: CLR,
                R1: INC,
        R2: LOOP,
    THEN,
    RET,
END-CODE

: EMIT      RS232-EMIT ;                        ( char -- )
: TYPE      RS232-TYPE ;                        ( $addr u -- )
: RTYPE     RS232-RTYPE ;                       ( addr u -- )
: CR        ^M RS232-EMIT  ^J RS232-EMIT ;      ( -- )
: SPACE     BL RS232-EMIT ;                     ( -- )
: SPACES    ?DUP IF  FOR  SPACE  NEXT  THEN ;   ( u -- )
: PAGE      ^L RS232-EMIT ;                     ( -- )

\ End
\ [PURPOSE]   Number output to current output device

VARIABLE BASE                           \ Hold number conversion base
VARIABLE HLD                            \ Work addr. in number conv. buffer
$80 CONSTANT NUM                        \ Start of number conversion buffer

\ Unsigned double devide routine

: M/MOD     >R  0  R@ UM/MOD  R> SWAP >R  UM/MOD R> ;   ( d1 u -- ur d2 )

\ Number conversion primitives

: HOLD      DECR HLD  FROM HLD ! ;                      ( char -- )
: <#        NUM TO HLD ;                                ( -- )
: #>        2DROP  FROM HLD  NUM FROM HLD - ;           ( d -- addr u )
: #         FROM BASE M/MOD  ROT 9 OVER <               ( d1 -- d2 )
            IF  7 +  THEN  [CHAR] 0 +  HOLD ;

\ Number output words

: DECIMAL   #10 TO BASE ;                               ( -- )
: HEX       $10 TO BASE ;                               ( -- )

\ End of number I/O

\ This clock is active when the DCF77 decoder is not synchronised.
CODE SPARE-CLOCK    ( -- )              \ Update clock when out of sync.
    SECONDS INC,                        \ Increase sec.
    A: SECONDS MOV,
    A: 60 # EQL IF,                     \ One minute passed ?
        SECONDS 0 # MOV,
        A: MINUTES MOV,                 \ Increase min.
        A: 1 # ADD,
        A: DA,                          \ Adjust BCD contents
        MINUTES A: MOV,
        A: $60 # EQL IF,                \ One hour passed !
            MINUTES 0 # MOV,
            A: HOURS MOV,               \ Increase hour
            A: 1 # ADD,
            A: DA,                      \ Adjust BCD contents
            HOURS A: MOV,
       THEN,
    THEN,
    RET,
END-CODE

\ Actives a clock when the DCF77 decoder is not synchronised.
CODE SPARE-TICK     ( -- )              \ Dcf Spare clock, sec. min. hr.
    A: STICK MOV,
    A: 100 # EQL IF,                    \ Ticked one second ?
        STICK 0 # MOV,                  \ Yes, reset spare tick
        ADR SYNC? BC IF,                \ Not synchonised ?
            ADR SEC? SETB,              \ Yes, one second passed !
            ADR LED2 CLR,               \ Spare second elapsed
            ' SPARE-CLOCK CALL,         \ Run spare clock !
        THEN,
    ELSE,
        A: 50 # EQL IF,                 \ 50 % Duty cycle for LED
            ADR LED2 SETB,
        THEN,
    THEN,
    RET,
END-CODE

\ Save separate data portions, from sixty serial DCF data bits.

CODE SAVE-INFO      ( -- )              \ Dcf Save INFO bits etc.
    R1: 21 # EQL IF,                    \ Low 5 bits valid
        A: RR,
        A: RR,
        A: $1F # ANL,                   \ Remove bit 20
        INFO: A: MOV,                   \ Info
    THEN,
    RET,
END-CODE

CODE SAVE-MINS      ( -- )
    R1: 29 # EQL IF,                    \ Low 7 bits valid
        A: $7F # ANL,                   \ Remove bit 28
        MINS: A: MOV,                   \ Minutes
    THEN,
    RET,
END-CODE

CODE SAVE-HOUR      ( -- )
    R1: 36 # EQL IF,                    \ Low 6 bits valid
        A: RR,
        A: $3F # ANL,                   \ Remove bit 35
        HOURS: A: MOV,                  \ Hours
    THEN,
    RET,
END-CODE

CODE SAVE-DAYS      ( -- )
    R1: 42 # EQL IF,                    \ Low 6 bits valid
        A: RR,
        A: RR,
        A: $3F # ANL,
        DAYS: A: MOV,                   \ Days
    THEN,
    RET,
END-CODE

CODE SAVE-WDAY      ( -- )
    R1: 45 # EQL IF,                    \ Low 3 bits valid
        A: SWAP,
        A: RR,
        A: 7 # ANL,
        WEEKDAY: A: MOV,                \ Weekday
    THEN,
    RET,
END-CODE

CODE SAVE-MONTH     ( -- )
    R1: 50 # EQL IF,                    \ Low 5 bits valid
        A: RR,
        A: RR,
        A: RR,
        A: $1F # ANL,
        MONTHS: A: MOV,                 \ Month
    THEN,
    RET,
END-CODE

CODE SAVE-YEAR      ( -- )
    R1: 58 # EQL IF,                    \ All 8 bits valid
        YEARS: A: MOV,                  \ Year
    THEN,
    RET,
END-CODE

CODE SAVE-WORK-REG  ( -- addr )         \ Dcf Save the work register
    A: WORK MOV,
    R1: BITS MOV,
    ' SAVE-INFO CALL,
    ' SAVE-MINS CALL,
    ' SAVE-HOUR CALL,
    ' SAVE-DAYS CALL,
    ' SAVE-WDAY CALL,
    ' SAVE-MONTH CALL,
    ' SAVE-YEAR JMP,
END-CODE

CODE COPY-DATA      ( -- addr )         \ Dcf Update user data each minute
    A: BITS MOV,
    A: 60 # EQL IF,                     \ One minute counted ?
        BITS 0 # MOV,                   \ Reset bit counter
        ADR SYNC? BS IF,                \ Still synchronised ?
            SECONDS 0 # MOV,            \ Start with zero again.
            MINUTES MINS: MOV,          \ Yes, copy all DCF77 data to user reg.
            HOURS HOURS: MOV,
            DAYS DAYS: MOV,
            MONTHS MONTHS: MOV,
            YEARS YEARS: MOV,
            WEEKDAY WEEKDAY: MOV,
            INFO INFO: MOV,
        THEN,
    THEN,
    RET,
END-CODE

CODE SYNCHRONISED   ( -- )              \ DCF signal synchronised ?
    A: TICK MOV,
    A: 50 # EQL IF,                     \ Led off after .5 sec.
        ADR LED1 SETB,
    THEN,
    A: 103 # EQL IF,                    \ More then one second passed ?
        A: BITS MOV,                    \ Yes,
        A: 58 # EQL IF,                 \ Is it the 58 second gap ?
            BITS INC,                   \ Yes, incr bit counter
            ADR SYNC? SETB,             \ Correct DCF77 message received
            SECONDS INC,                \ Yes, increase second 59
            ADR SEC? SETB,              \ Second passed !
        ELSE,
            ADR SYNC? CLR,              \ No longer synchronised
        THEN,
    ELSE,
        A: 105 # EQL IF,                \ Not sync. and in 58 sec. gap ?
            BITS 59 # MOV,              \ Yes, setup bit counter again
        THEN,
    THEN,
    RET,
END-CODE

CODE READ-BIT       ( -- )              \ Read next DCF77 bit
    A: TICK MOV,                        \ Get clock ticks
    A: 15 # EQL IF,                     \ 150 MS passed ?
        C: ADR DCF-IN MOV,              \ Yes, High = 1 (180ms> low pulse)
        A: WORK MOV,                    \ Low = 0       (120ms< low pulse)
        C: CPL,                         \ Invert bit
        ADR BIT C: MOV,                 \ Save last received DCF77 bit
        A: RRC,                         \ Shift bit into work register
        WORK A: MOV,
    THEN,
    RET,
END-CODE

\G Timer interrupt routine, increases one hundred times a second
\G a 8 bits timer. It checks for the DCF77 pulses to come in
\G exactly every second (+ or - 2%). Second 59 is generated
\G automaticly. In conjunction with the DCF interrupt it controls
\G if the DCF77 signal is correct received. It will decode low and
\G high input bits, from the DCF77 signal and shift these into the
\G work register.
CODE TIMER          ( -- addr )         \ Dcf Executed every PERIOD ticks
    PSW: PUSH,                          \ To synchronise on DCF77 signal
    ACC: PUSH,
    TH0: HPERIOD # MOV,                 \ Set negated 16 bit count (10 ms)
    TL0: LPERIOD # MOV,
    TICK INC,                           \ Increase timer clock
    STICK INC,                          \ Increase spare timer clock
    ' SYNCHRONISED CALL,                \ Check bitstream
    ' READ-BIT CALL,                    \ Read next bit
    ' SPARE-TICK CALL,                  \ Spare clock
    ' COPY-DATA CALL,                   \ Copy user data each minute
    ACC: POP,
    PSW: POP,
    RETI,                               \ Return
END-CODE

\G EX4 interrupt routine, executed on every DCF77 second pulse.
\G It resets the TICK timer. Saves the result of the timer interrupt
\G in a test location. Checks for a correct DCF77 second pulse.
\G It saves the work register in the data array after the data of
\G a message type is completed. When a new minute starts the 
\G result is copied to the user registers.
CODE DCF            ( -- )              \ Dcf Executed once every second
    PSW: PUSH,
    ACC: PUSH,
    [R1] PUSH,
    A: TICK MOV,                        \ Valid DCF77 second pulse ?
    C: CLR,
    A: 98 # SUBB,
    CC IF,
        TICK 0 # MOV,                   \ Yes, reset tick (1/100 sec) counter
        ADR SYNC? BS IF,                \ Only when synchronised
            STICK 0 # MOV,              \ Reset spare tick
        THEN,
        ADR LED1 CLR, 
        BITS INC,                       \ Increase bit counter
        ADR SYNC? BS IF,                \ Synchronised ?
            SECONDS INC,                \ Yes, incr. second counter
            ADR SEC? SETB,              \ Next second !
        THEN,
        ' SAVE-WORK-REG CALL,           \ Save data from work register
    THEN,
    [R1] POP,
    ACC: POP,
    PSW: POP,
    RETI,
END-CODE

\G Start DCF77 decoder and thereby timer 0, reset timer and 
\G enable timer interrupts. Enable IE1 interrupts from P3.3
CODE SETUP-DCF      ( -- )              \ Dcf Setup decoder hardware
    ADR SYNC? CLR,                      \ Start out of synchronisation
    ADR SEC? CLR,                       \ Show first second always
    TCON: .2 SETB,                      \ EX1 active on falling edges
    IE: .2 SETB,                        \ EX1 interrupt on
    ADR DCF-IN SETB,                    \ Use P3.3 for input
    TMOD: $F0 # ANL,                    \ Clear Timer 0 mode bits
    TMOD: $01 # ORL,                    \ Set Timer 0 mode 1
    TH0: HPERIOD # MOV,                 \ Set negated 16 bits count (10ms)
    TL0: LPERIOD # MOV,
    EAL: SETB,                          \ Global enable interrupts
    ET0: SETB,                          \ Enable timer 0 interrupts
    TR0: SETB,                          \ Start timer 0
    A: CLR,                             \ Clear DCF77 clock registers
    BITS A: MOV,                        \ Bit counter
    SECONDS A: MOV,                     \ Second counter
    TICK A: MOV,                        \ Tick counter
    STICK A: MOV,                       \ Spare tick
    RET,
END-CODE

\ High level part of DCF77 primitives

: .BCD              ( byte -- )         \ Dcf Print BCD numbers
    HEX  0 <# # # #> RTYPE  DECIMAL
    ;

: .SEC              ( +n -- )           \ Dcf Print seconds
    0 <# # # #> RTYPE SPACE
    ;

: .SYNC             ( -- )              \ Dcf DCF77 decoder synchronised
\   FROM BITS 0 <# # # #> RTYPE SPACE   \ Show incoming DCF77 bit count
\   BIT .BCD SPACE                      \ and latest bit.
    SYNC? IF  ." +DCF: "  ELSE  ." -DCF: "  THEN
    ;

: .DASH             ( -- )              \ Dcf print dash
    &- EMIT
    ;

: .SUMMER           ( -- )              \ Dcf Show summer/winter time mark
    FROM INFO 4 AND IF  ." Zom"  ELSE  ." Wint"  THEN  ." ertijd "
    ;

: .DAY              ( -- )              \ Dcf Print weekday
    FROM WEEKDAY
    BEGIN-SELECT
        SELECT 1   ." Maandag "   EXIT
        SELECT 2   ." Dinsdag "   EXIT
        SELECT 3   ." Woensdag "  EXIT
        SELECT 4   ." Donderdag " EXIT
        SELECT 5   ." Vrijdag "   EXIT
        SELECT 6   ." Zaterdag "  EXIT
        SELECT 7   ." Zondag "    EXIT
    END-SELECT
    ;

: .DATE             ( -- )              \ Dcf Print DCF77 date data
    ." Datum: "  .DAY
    FROM DAYS    .BCD   .DASH           \ .BCD is handy for printing BCD data
    FROM MONTHS  .BCD   .DASH
    FROM YEARS   .BCD   SPACE
    ;

: .TIME             ( -- )              \ Dcf Print DCF77 time data
    ." Klok: "
    FROM HOURS      .BCD   .DASH
    FROM MINUTES    .BCD   .DASH
    FROM SECONDS    .SEC                \ Seconds are normal binary data
    ;

: DCF-CLOCK     ( -- )
    SETUP                               \ Install Forth machine
    SETUP-RS232                         \ Setup RS232 port
    SETUP-DCF                           \ Initialise hardware & registers
    DECIMAL                             \ Number base decimal
    PAGE ." DCF77 decoder vsn 1.00"     \ Startup message
    CLEAR LED2                          \ Led 2 on
    50 FOR  100 MS  NEXT                \ for 2.5 to 5 sec.
    SET LED2                            \ Led 2 off
    CR ." Synchronizing "
    BEGIN 
        SEC? IF 
            CLEAR SEC?
            &. EMIT                     \ Show advance
        THEN
    SYNC? UNTIL
    BEGIN
        SEC? IF                         \ One sec. passed,
            CLEAR SEC?                  \ Yes,
            CR                          \ Send time to RS232 port
            .SYNC                       \ Print synchroniser data
            .SUMMER                     \ Print summer/wintertime mark
            .DATE                       \ Print current date
            .TIME                       \ Print current time
        THEN
    AGAIN
    ;

\ Install the DCF77 decoder interrupts, and start decoder.

' TIMER     TF0-VEC   SET-VECTOR        \ Install timer 0 interrupt
' DCF       IE1-VEC   SET-VECTOR        \ Install external interrupt 1
' DCF-CLOCK RESET-VEC SET-VECTOR        \ Install program

                    \ (* End of Source *) /
