\ ==============================================================
\ [TITLE]     AT89Cx051 ByteForth Hitachi LCD library. 
\ [FILE]      LCD.FRT
\ [CATEGORY]  Library ( 357 bytes )
\ [VSN]       1.11 
\ [CREATED]   April 9, 1995, Willem Ouwerkerk 
\ [LASTCHNGD] June 15, 1997, Willem Ouwerkerk 
\ [COPYRIGHT] De Elektrise Schaar (C) 1995-1997
\ [USES]      ByteForth cross compiler
\ [PURPOSE]   LCD-display thru port P1 of the AT89C2051
\             Use a display with the Hitachi HD44780A00H chip,
\             and 4 bits LCD I/O communication (6MHz). 
\ ==============================================================

NEEDS TARGET.FRT

ATOM INLINE$        ATOM MS             \ Import these macros

\ This method is not valid for a four line UC-16401 display.
\ CR and AT-XY  does not work properly. This because the
\ display has only two lines internally. These lines are
\ interlaced on the four display lines. 

\ LCD-display dimensions and cursor position

40 CONSTANT C/L                         \ Screen width
2 CONSTANT L/S                          \ Lines per screen
VARIABLE OUT                            \ Current cursor position
FLAG DATA?                              \ Lcd data type, true is CHAR

ASSEMBLER
P1: CONSTANT PORT                       \ I/O port for LCD using 6 bits
FORTH

\ Send one data nibble (4 bits) x to the LC-display. When flag
\ is true x is accepted as data. When flag is false x is accepted
\ as instruction.
CODE NIBBLE-TO-LCD  ( nibble -- )       \ Lcd
    A: @R0 MOV,
    R0: INC,
    A: $0F # ANL,                       \ Clear upper four bits
    C: ADR DATA? MOV,
    ACC: .5 C: MOV,                     \ Bit P1.5 is RS
    PORT A: MOV,
    PORT .4 SETB,                       \ Bit P1.4 is E
    PORT .4 CLR,
    RET,
END-CODE

\ Send one character char as two nibbles to a LCD-display. When
\ DATA? is true, char is accepted as data. When DATA? is false,
\ char is accepted as instruction.

\G Stuur een instructie byte naar een LC-display.
: LCD-CHAR      ( byte -- )             \ Lcd
    SPLIT  NIBBLE-TO-LCD
           NIBBLE-TO-LCD
    ;

\G Stuur een karakter char naar een LC-display.
: LCD-INSTR      ( char -- )            \ Lcd
    CLEAR DATA?
    LCD-CHAR
    SET DATA?  
    ;

\G Maak het LC-display schoon en zet de cursor in linker bovenhoek.
: LCD-PAGE      ( -- )                  \ Lcd Clean LCD and restore cursor
    1 LCD-INSTR  2 MS
    CLEAR OUT
    ;

\G Initialiseer een LC-display. Zet 4 bits data bus, maak het 
\G display schoon, zet de cursor uit en zet de display richting 
\G van links naar rechts.
: LCD-INIT      ( -- )                  \ Lcd
    CLEAR DATA?                         \ Initialise LCD controller
    16 MS                               \ Wait and set 4 bits interface mode
    3 FOR
        %0011 NIBBLE-TO-LCD  5 MS
    NEXT
    %0010 NIBBLE-TO-LCD                 \ Four bits interface set
    $28 LCD-INSTR                       \ Four bits interf. two line display
    $08 LCD-INSTR                       \ Display off
    LCD-PAGE                            \ Clear display
    $06 LCD-INSTR                       \ Shift right
    $0C LCD-INSTR                       \ Display on & cursor off
    ;

\G Zet een karakter char op het LC-display.
: LCD-EMIT      ( char -- )             \ Lcd Output one char
    LCD-CHAR  INCR OUT
    ;

\G Zet u karakters vanaf het ROM adres $addr naar het LC-display.
: LCD-TYPE      ( $addr u -- )          \ Lcd Output u chars from $addr
    ?DUP 
    IF      FOR  COUNT LCD-EMIT  NEXT  
    THEN    2DROP
    ;

\G Zet u karakters vanaf het RAM adres addr naar een LC-display.
: LCD-RTYPE     ( addr u -- )           \ Lcd Output u chars from addr
    ?DUP 
    IF      FOR  @+ LCD-EMIT  NEXT  
    THEN    DROP
    ;

\G Zet de cursor in de linkerbovenhoek van het LC-display.
: LCD-HOME      ( -- )                  \ Lcd
    $80 LCD-INSTR                       \ Put cursor in upper left corner of LCD
    CLEAR OUT
    ;

\G Wis het vorige karakter van het LC-display.
: LCD-BS        ( -- )                  \ Lcd Delete previous character if any
    OUT
    IF      $10 LCD-INSTR
            BL LCD-CHAR
            $10 LCD-INSTR
            DECR OUT
    THEN
    ;

\G Zet een spatie op het LC-display.
: LCD-SPACE     ( -- )                  \ Lcd Print one space on a LCD-display
    BL LCD-EMIT
    ;

\G Zet u spaties op het LC-display.
: LCD-SPACES    ( u -- )                \ Lcd Print u spaces on a LCD-display
    ?DUP 
    IF      FOR  LCD-SPACE  NEXT
    THEN
    ;

\G Zet de cursor op position x y van het LC-display.
\G Let op: Dit werkt zo alleen op 40 kar. displays.
: LCD-AT-XY     ( x y -- )              \ Lcd Put cursor at pos x of line y
    LCD-HOME
    C/L * +  ?DUP
    IF      DUP TO OUT
            FOR  $10 LCD-INSTR  NEXT
    THEN
    ;

\ Does not work for a four line display like UC-16401

\G Doe een CR op een LC-display.
\G Let op: Dit werkt zo alleen op 40 kar. displays.
: LCD-CR        ( -- )                  \ Lcd Advance cursor to a new line
    C/L  FROM OUT  -  ?DUP
    IF  LCD-SPACES  THEN
    FROM OUT  C/L L/S * 1- AND  TO OUT
    ;

\ Generic output routines

: EMIT      LCD-CHAR ;              ( char -- )
: TYPE      LCD-TYPE ;              ( $addr u -- )
: RTYPE     LCD-RTYPE ;             ( addr u -- )
: CR        LCD-CR ;                ( -- )
: SPACE     LCD-SPACE ;             ( -- )
: SPACES    LCD-SPACES ;            ( u -- )
: PAGE      LCD-PAGE ;              ( -- )
: BS        LCD-BS ;                ( -- )

DOC  LCD demo

: MAIN          ( n -- )                \ Display character set on LCD
    SETUP
    LCD-INIT
    BEGIN
        0 FOR
            255 I - EMIT  50 MS
        NEXT
    AGAIN
    ;   

' MAIN      RESET-VEC  SET-VECTOR

ENDDOC

\ End
